import dataclasses
import uuid

from flask import Flask, make_response, render_template, request, session

from zndraw import globals, io

app = Flask(__name__)
app.secret_key = str(uuid.uuid4())


@app.route("/")
def index():
    session["key"] = str(uuid.uuid4())  # TODO use session key e.g. for atoms cache
    return render_template("index.html", config=dataclasses.asdict(globals.config))


@app.route("/config")
def config():
    return {
        **dataclasses.asdict(globals.config),
        "total_frames": len(globals._atoms_cache) - 1,
    }


@app.route("/atoms", methods=["POST"])
def atoms_step():
    step = request.json
    try:
        atoms = globals.config.get_atoms(step=int(step))
        graph = io.get_graph(atoms)
        return [{**graph.nodes[idx], "id": idx} for idx in graph.nodes]
    except (KeyError, IndexError):
        return []


@app.route("/positions", methods=["POST"])
def positions_step():
    params = request.json
    result = []
    try:
        for step in range(params["start"], params["stop"]):
            atoms = globals.config.get_atoms(step=int(step))
            result.append(atoms.get_positions().tolist())
        return result
    except (KeyError, IndexError):
        return result


@app.route("/bonds", methods=["POST"])
def bonds_step():
    step = request.json
    atoms = globals.config.get_atoms(step=int(step))
    graph = io.get_graph(atoms)
    return list(graph.edges)


@app.route("/select", methods=["POST"])
def select() -> list[int]:
    """Update the selected atoms."""
    step = request.json["step"]
    selected_ids = request.json["selected_ids"]
    return {"selected_ids": selected_ids, "updated": False}

    # atoms = globals.config.get_atoms(step)

    # for id in tuple(selected_ids):
    #     selected_symbol = atoms[id].symbol
    #     selected_ids += [
    #         idx for idx, atom in enumerate(atoms) if atom.symbol == selected_symbol
    #     ]

    # return {"selected_ids": list(set(selected_ids)), "updated": True}


@app.route("/add_update_function", methods=["POST"])
def add_update_function():
    """Add a function to the config."""
    globals.config.update_function = request.json
    try:
        signature = globals.config.get_update_signature()
    except (ImportError, ValueError) as err:
        return {"error": str(err)}
    return signature


@app.route("/update_function_values", methods=["POST"])
def update_function_values():
    """Update the values of the update function."""
    globals.config.set_update_function_parameters(request.json)
    return {}


@app.route("/select_update_function/<name>")
def select_update_function(name):
    """Select a function from the config."""
    globals.config.update_function_name = name
    return {}


@app.route("/update", methods=["POST"])
def update_scene():
    selected_ids = list(sorted(request.json["selected_ids"]))
    step = request.json["step"]

    function = globals.config.get_update_function()
    atoms = function(
        [int(x) for x in selected_ids], globals.config.get_atoms(step=int(step))
    )

    offset = len(globals._atoms_cache)

    for idx, atom in enumerate(atoms):
        print(f"processing {idx}")
        globals._atoms_cache[idx + offset] = atom

    # this has to return before the scene is automatically updated
    return {}


@app.route("/load")
def load():
    """Function to call asynchronously to load atoms in the background."""
    globals.config.load_atoms()
    return {}



import contextlib
import pathlib
import socket
import sys
import webbrowser

import typer

from zndraw import __version__, app, globals

try:
    import webview as wv
except ImportError:
    wv = None

cli = typer.Typer()


def version_callback():
    typer.echo(f"ZnDraw {__version__}")
    raise typer.Exit()


@cli.command()
def main(
    file: str = typer.Argument(..., help="Trajectory File"),
    port: int = typer.Option(None, help="Port to run the server on"),
    animate: bool = typer.Option(False, help="Animate the trajectory"),
    sphere_size: float = typer.Option(1.0, help="size of the hydrogen sphere"),
    bond_size: float = typer.Option(1.0, help="size of a bond"),
    max_fps: int = typer.Option(100, help="Maximum frames per second"),
    update_function: str = typer.Option(
        None, help="Path to a python file with an update function 'module.function'."
    ),
    repeat: int = typer.Option(1, help="Repeat the trajectory n times"),
    resolution: int = typer.Option(5, help="Proportional to the number of polygons."),
    restart_animation: bool = typer.Option(
        False, help="run the animation in an endless loop."
    ),
    frames_per_post: int = typer.Option(100, help="Number of frames to send per POST."),
    browser: bool = typer.Option(True, help="Open the browser automatically."),
    webview: bool = typer.Option(True, help="Use the webview library if available."),
):
    """ZnDraw: Visualize Molecules

    The ZnDraw CLI. Use 'zndraw version' to get the current version.
    """
    sys.path.insert(1, pathlib.Path.cwd().as_posix())
    if port is None:
        sock = socket.socket()
        sock.bind(("", 0))
        port = sock.getsockname()[1]
        sock.close()

    if file == "version":
        version_callback()

    if not pathlib.Path(file).exists():
        typer.echo(f"File {file} does not exist.")
        raise typer.Exit()

    globals.config.file = file
    globals.config.animate = animate
    globals.config.sphere_size = sphere_size
    globals.config.bond_size = bond_size
    globals.config.max_fps = max_fps
    globals.config.update_function = update_function
    globals.config.resolution = resolution
    globals.config.frames_per_post = frames_per_post
    globals.config.restart_animation = restart_animation
    globals.config.repeat = (repeat, repeat, repeat)

    if wv is not None and webview:
        wv.create_window("ZnDraw", app)
        with contextlib.suppress(wv.WebViewException):
            wv.start()
            return
    if browser:
        webbrowser.open(f"http://localhost:{port}")
    app.run(port=port)



import dataclasses
import importlib
import pathlib

import ase.io
import pydantic
import tqdm
import znh5md


@dataclasses.dataclass
class Config:
    file: str = None
    animate: bool = None
    sphere_size: float = None
    bond_size: float = None
    max_fps: int = None
    update_function: str = None
    frames_per_post: int = 100
    restart_animation: bool = False
    resolution: int = 5
    repeat: tuple = (1, 1, 1)

    _update_function_name: str = None

    @property
    def update_function_name(self):
        if self._update_function_name is not None:
            return self._update_function_name
        return self.update_function.rsplit(".", 1)[1]

    @update_function_name.setter
    def update_function_name(self, value):
        self._update_function_name = value

    def get_update_signature(self):
        module_name, function_name = self.update_function.rsplit(".", 1)
        if function_name in _update_functions:
            return _update_functions[self.update_function_name].schema()
        module = importlib.import_module(module_name)
        instance: pydantic.BaseModel = getattr(module, function_name)()
        _update_functions[function_name] = instance
        return _update_functions[self.update_function_name].schema()

    def get_update_function(self):
        module_name, function_name = self.update_function.rsplit(".", 1)
        if self.update_function is None:
            return None
        if function_name in _update_functions:
            return _update_functions[self.update_function_name].run

        module = importlib.import_module(module_name)
        _update_functions[self.update_function_name] = getattr(module, function_name)()
        return _update_functions[self.update_function_name].run

    def set_update_function_parameters(self, value):
        instance = _update_functions[value["function_id"]]
        attribute = value["property"].lower()
        value = value["value"]
        if instance.__annotations__[attribute] == float:
            value = float(value)
        elif instance.__annotations__[attribute] == int:
            value = int(value)
        elif instance.__annotations__[attribute] == bool:
            value = bool(value)
        else:
            value = value
        print(f"Setting {attribute} of {instance} to {value}")
        setattr(instance, attribute, value)

    def load_atoms(self, item=None):
        if item == 0:
            if pathlib.Path(self.file).suffix == ".h5":
                _atoms_cache[0] = znh5md.ASEH5MD(self.file)[0]
            else:
                _atoms_cache[0] = ase.io.read(self.file)
        elif self.update_function is not None:
            return
        else:
            if pathlib.Path(self.file).suffix == ".h5":
                _atoms_cache.update(
                    dict(enumerate(znh5md.ASEH5MD(self.file).get_atoms_list()))
                )
            else:
                for idx, atom in enumerate(
                    tqdm.tqdm(ase.io.iread(self.file), desc="File Reading")
                ):
                    _atoms_cache[idx] = atom

    def get_atoms(self, step=0) -> ase.Atoms:
        try:
            return _atoms_cache[step].repeat(self.repeat)
        except KeyError:
            if step != 0:
                raise
            self.load_atoms(0)
            return _atoms_cache[0]


# TODO set defaults here and load in typer?

_update_functions = {}

_atoms_cache: dict = {}
config = Config()

graph = None
atoms = None



import abc

import ase
import numpy as np
from pydantic import BaseModel, Field


class UpdateScene(BaseModel, abc.ABC):
    @abc.abstractmethod
    def run(self, atom_ids: list[int], atoms: ase.Atoms) -> list[ase.Atoms]:
        pass


class Explode(UpdateScene):
    steps: int = Field(100, le=1000, ge=1)
    particles: int = Field(10, le=100, ge=1)

    def run(self, atom_ids: list[int], atoms: ase.Atoms) -> list[ase.Atoms]:
        particles = []
        for _atom_id in atom_ids:
            for _ in range(self.particles):
                particles.append(ase.Atoms("Na", positions=[atoms.positions[_atom_id]]))

        for _ in range(self.steps):
            struct = atoms.copy()
            for particle in particles:
                particle.positions += np.random.normal(scale=0.1, size=(1, 3))
                struct += particle
            yield struct


class Delete(UpdateScene):
    def run(self, atom_ids: list[int], atoms: ase.Atoms) -> list[ase.Atoms]:
        for idx, atom_id in enumerate(sorted(atom_ids)):
            atoms.pop(atom_id - idx)  # we remove the atom and shift the index
        return [atoms]


class Move(UpdateScene):
    x: float = Field(0.5, le=5, ge=0)
    y: float = Field(0.5, le=5, ge=0)
    z: float = Field(0.5, le=5, ge=0)

    def run(self, atom_ids: list[int], atoms: ase.Atoms) -> list[ase.Atoms]:
        for atom_id in atom_ids:
            atom = atoms[atom_id]
            atom.position += np.array([self.x, self.y, self.z])
            atoms += atom
        return [atoms]


class Duplicate(UpdateScene):
    x: float = Field(0.5, le=5, ge=0)
    y: float = Field(0.5, le=5, ge=0)
    z: float = Field(0.5, le=5, ge=0)
    symbol: str = Field("same")

    def run(self, atom_ids: list[int], atoms: ase.Atoms) -> list[ase.Atoms]:
        for atom_id in atom_ids:
            atom = ase.Atom(atoms[atom_id].symbol, atoms[atom_id].position)
            atom.position += np.array([self.x, self.y, self.z])
            atom.symbol = self.symbol if self.symbol != "same" else atom.symbol
            atoms += atom
        return [atoms]


class ChangeType(UpdateScene):
    symbol: str = Field("")

    def run(self, atom_ids: list[int], atoms: ase.Atoms) -> list[ase.Atoms]:
        for atom_id in atom_ids:
            atoms[atom_id].symbol = self.symbol
        print(atoms)
        return [atoms]
